class OR(object):
    """Specify that a value should be one of several types.
    """

    def __init__(self, *args):
        self.valid_types = tuple(sorted(args))

    def evaluate(self, value):
        """Determine if :param value: meets this operator's criteria."""
        return isinstance(value, self.valid_types)

    def __eq__(self, other):
        """An equality predicate to simplify unit testing."""
        if not hasattr(other, 'valid_types'):
            return False
        return self.valid_types == other.valid_types

    def __repr__(self):
        return '<OR {0}>'.format(', '.join([repr(x) for x in self.valid_types]))


class IS(object):
    """Specify that a value should one of several *exact values*.

    This operator uses a ``set`` for fast checking so any values must be
    hashable.
    """

    def __init__(self, *args):
        self.valid_values = set(args)

    def evaluate(self, value):
        """Determine if :param value: meets this operator's criteria."""
        return value in self.valid_values

    def __eq__(self, other):
        """An equality predicate to simplify unit testing."""
        if not hasattr(other, 'valid_values'):
            return False
        return self.valid_values == other.valid_values

    def __repr__(self):
        return '<IS {0}>'.format(', '.join([repr(x) for x in self.valid_values]))



# -*- coding: utf-8 -*-


class Cursor(object):
    """Wrapper for :class:`pymongo.cursor.Cursor` objects.

    This class is a thin wrapper which takes results returned by the underlying
    cursor and wraps them in the appropriate
    :class:`~scalymongo.document.Document` subclass.
    """

    def __init__(self, wrapped_cursor, document_type):
        self.__wrapped_cursor = wrapped_cursor
        self.__document_type = document_type

    def __getitem__(self, index):
        """Get the item(s) at `index`.

        :param index: may also be a slice.
        """
        returned = self.__wrapped_cursor[index]
        # If the index is a slice then the result is a new cursor with the skip
        # and limit already applied.
        if isinstance(index, slice):
            return Cursor(returned, self.__document_type)
        return self.__document_type(returned)

    def next(self):
        """Return the next document for this cursor.
        """
        return self.__document_type(self.__wrapped_cursor.next())

    def clone(self):
        """Return a new wrapper :class:`Cursor`.
        """
        return Cursor(self.__wrapped_cursor.clone(), self.__document_type)

    def __getattr__(self, attr):
        """All other methods and properties are those of the wrapped cursor.
        """
        return getattr(self.__wrapped_cursor, attr)



"""
Document
========

The base document models.
"""
import functools

from pymongo.errors import OperationFailure

from scalymongo.cursor import Cursor
from scalymongo.errors import (
    GlobalQueryException,
    ModifyFailedError,
    UnsafeBehaviorError,
)
from scalymongo.helpers import is_update_modifier, value_or_result
from scalymongo.schema import (
    SchemaDocument,
    SchemaMetaclass,
    validate_update_modifier,
)


class DocumentMetaclass(SchemaMetaclass):
    """This metaclass records all concrete :class:`Document` types.

    This excludes classes created by
    :meth:`~scalymongo.connection.Connection.connect_document`.
    """

    concrete_classes = set()

    def __new__(cls, name, bases, attrs):
        rv = SchemaMetaclass.__new__(cls, name, bases, attrs)
        attrs['abstract'] = attrs.get('abstract', False)
        if not attrs['abstract'] and not attrs.get('connection'):
            cls.concrete_classes.add(rv)
        return rv


def get_concrete_classes():
    """Return a set of all non-abstract :class:`Document` subclasses."""
    return DocumentMetaclass.concrete_classes


class Document(SchemaDocument):
    """The base document class all user models should extend.

    Any subclasses of :class:`Document` which are not abstract can be used from
    the model property on a :class:`~scalymongo.connection.Connection`.  This
    class wraps many of the methods available on
    :class:`pymongo.collection.Collection` in order to return objects of the
    appropriate subclass.
    """

    __metaclass__ = DocumentMetaclass
    abstract = True
    default_values = {}
    """A dictionary mapping fields to default values.

    The values may either be static values or a function that returns a default
    value.  For example to put a new :class:`~uuid.UUID` in the field `uuid` you
    would use a default like:

    .. code-block:: python

        default_values = {'uuid': uuid.UUID}
    """

    def __init__(self, *args, **kwargs):
        SchemaDocument.__init__(self, *args, **kwargs)
        for key, value in self.default_values.iteritems():
            self.setdefault(key, value_or_result(value))

    def save(self):
        if '_id' in self:
            raise UnsafeBehaviorError(
                'This document has already been saved once.'
                ' Further alterations should use modify.')

        self.validate()
        self.collection.insert(self)

    def reload(self):
        """Reload this document.

        This will make use of the shard key for sharded collections to avoid a
        global query.
        """
        spec = self.shard_key
        spec['_id'] = self['_id']
        # Call the parent `update` not the classmethod.
        SchemaDocument.update(self, self.find_one(spec))

    def modify(self, update, query=None):
        """Modify this document using :meth:`find_and_modify`.

        If the document could not be updated a
        :class:`~scalymongo.errors.ModifyFailedError` is raised.

        :param update: Is an update modifier or replacement document to be
            be used for this

        :param query: A query specification for any additional parameters to
            be used in the :meth:`find_and_modify` query.  This document's
            ``_id`` field and it's shard key fields (where applicable) are
            included in addition to any parameters specified in the
            `query`.
        """
        full_query = self.shard_key
        if query:
            full_query.update(query)
        full_query['_id'] = self['_id']
        result = self.find_and_modify(full_query, update, new=True)
        if result is None:
            self.reload()
            raise ModifyFailedError(
                'Failed to update document.  The document was not found'
                ' based on the criteria {0}.  Document was {1}.'.format(
                query, self),
            )
        # We have to clear the existing dictionary first in case the operation
        # included an `$unset` or replaced the entire document.
        self.clear()
        # Call the parent `update` not the classmethod.
        SchemaDocument.update(self, result)

    @classmethod
    def ensure_indexes(cls, **kwargs):
        """Ensure this any indexes declared on this index :class:`Document`.

        This is an administrative task and should be done with care as
        (re)building large indexes can make a database unusable for some time.
        Any additional `kwargs` are passed to
        :meth:`pymongo.collection.Collection.ensure_index`.
        """
        for index in cls.indexes:
            kwargs['unique'] = index.get('unique', False)
            cls.collection.ensure_index(index['fields'], **kwargs)

    @classmethod
    def find_one(cls, spec=None, allow_global=False, **kwargs):
        """Find and return one matching document of this type.

        :param spec: (optional): Is a query to find the matching document.
          Note that unlike other tools providing `find_one` scalymongo requires
          this value to be a dictionary including at a bare minimum the shard
          key for this models collection.

        :param kwargs: (optional): Additional keyword arguments will be
          passed to :meth:`pymongo.collection.Collection.find_one`.
        """
        if not allow_global:
            cls.check_query_sharding(spec)
        result = cls.collection.find_one(spec, **kwargs)
        if result is not None:
            return cls(result)

    @classmethod
    def find(cls, spec=None, allow_global=False, *args, **kwargs):
        """Query the database for documents of this type.

        This is a wrapped call to :meth:`pymongo.collection.Collection.find`
        which returns document objects of the appropriate class instead of
        ``dict`` instances.  All additional arguments are passed to the
        underlying method.
        """
        if not allow_global:
            cls.check_query_sharding(spec)
        result = cls.collection.find(spec, *args, **kwargs)
        return Cursor(result, cls)

    @classmethod
    def find_and_modify(cls, query={}, update=None,
                        allow_global=False, **kwargs):
        """Find and atomically update a single document.

        .. note::
            This method returns the *old* (pre-update) version of the document
            by default.  This is the default for ``findAndModify`` in the Mongo
            shell and it is maintained here for the sake of consistency.  Pass
            `new` as ``True`` to return the updated document instead.

        :param query: filter for the update (default ``{}``)
        :param update: see second argument to :meth:`update` (no default)
        :keyword allow_global: If ``True`` the query will not be checked to
            ensure that it contains the full shard key.  (default ``False``)
        :param sort: priority if multiple objects match (default ``{}``)
        :param remove: remove rather than updating (default ``False``)
        :param new: return updated rather than original object
            (default ``False``)
        :param fields: see second argument to :meth:`find` (default all)
        :param upsert: insert if object doesn't exist (default ``False``)
        :param kwargs: any other options the findAndModify_ command
            supports can be passed here.

        .. _findAndModify: http://www.mongodb.org/display/DOCS/findAndModify+Command
        """
        if not allow_global:
            cls.check_query_sharding(query)

        cls._validate_update(update)

        try:
            # Use the `command` operation since `find_and_modify` is only
            # available in pymongo>=1.10.
            returned = cls.database.command(
                'findandmodify', cls.collection.name,
                query=query, update=update, **kwargs)
        except OperationFailure:
            return None

        if returned['value'] is None:
            return None

        return cls(returned['value'])

    @classmethod
    def update(cls, spec, document, allow_global=False, **kwargs):
        """Update a document matching `spec` using `document`.

        Additional keyword arguments are passed to
        :meth:`pymongo.collection.Collection.update`.

        :param spec: is a document specification (i.e. query) for the document
            to be updated.  if more than one document matches only the first
            will but updated unless `multi` is ``True``.
        :param document: is expected to be either a new document or an update
            modifier.
        :keyword multi: update multiple documents or just the first (default
            ``False``).
        """
        if not allow_global:
            cls.check_query_sharding(spec)

        cls._validate_update(document)

        return cls.collection.update(spec, document, **kwargs)

    @classmethod
    def remove(cls, spec, allow_global=False, **kwargs):
        """Find and remove documents matching `spec`.

        Additional keywords are passed to
        :meth:`pymongo.collection.Collection.remove`.
        """
        if not allow_global:
            cls.check_query_sharding(spec)

        return cls.collection.remove(spec, **kwargs)

    @classmethod
    def _validate_update(cls, document):
        """Validate an update described in `document`.

        :param document: is expected to be either a new document or an update
            modifier.
        """
        if is_update_modifier(document):
            validate_update_modifier(document, cls.structure)
        else:
            # It's a full document replace.
            cls(document).validate()

    @classmethod
    def check_query_sharding(cls, spec):
        """Check that all required keys are present in `spec`.

        If any shard keys are unspecified this will raise a
        :class:`GlobalQueryException`.
        """
        if cls.shard_index:
            keys = set(cls.shard_index['fields'])
            missing_keys = keys.difference(spec.keys())
            if missing_keys:
                raise GlobalQueryException(
                    'Some or all of the shard key was not specified.  Missing'
                    ' fields were {0}.'.format(
                        ', '.join([key for key in missing_keys])))

    @property
    def shard_key(self):
        if not self.shard_index:
            return {}
        key_dict = {}
        for key in self.shard_index['fields']:
            key_dict[key] = self[key]
        return key_dict



#!/usr/bin/env python

import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='scalymongo',
    version='0.1.3',
    description='A scaling-centric MongoDB object document mapper',
    keywords = 'mongo sharding db',
    url='https://github.com/allancaffee/scaly-mongo',
    author='Allan Caffee',
    author_email='allan.caffee@gmail.com',
    license='BSD',
    packages=['scalymongo'],
    install_requires=['pymongo>=1.9'],
    test_suite='tests',
    long_description=read('README.rst'),
    zip_safe=True,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Database',
    ],
)



"""
Schema
======

The core schema validation logic.
"""

from scalymongo.errors import SchemaError, ValidationError
from scalymongo.helpers import ConversionDict
from scalymongo.structure_walker import StructureWalker


class UpdatingList(list):
    """Provide a list with an update method.

    Since dicts are used to describe some attributes they must use a data
    structure that does not rely on hashing.  This wrapper class just tacks the
    :meth:`update` method onto a list to make them interoperable with ``set``
    and ``dict`` instances.
    """
    update = list.extend


class SchemaMetaclass(type):
    """Metaclass for documents that have a schema."""

    mergeable_attrs = {
        'structure': dict,
        'indexes': UpdatingList,
        'required_fields': set,
        'default_values': dict,
    }
    "Map the merged base class attributes to their expected types."

    def __new__(cls, name, bases, attrs):
        for key, type_ in cls.mergeable_attrs.iteritems():
            if key in attrs:
                attrs[key] = type_(attrs[key])
            else:
                attrs[key] = type_()

        for base in bases:
            for field, type_ in cls.mergeable_attrs.iteritems():
                if hasattr(base, field):
                    attrs[field].update(getattr(base, field))

        attrs['shard_index'] = find_shard_index(attrs['indexes'])

        if name != 'SchemaDocument' and 'structure' in attrs:
            attrs['_conversions'] = make_conversion_dict(attrs['structure'])

        return type.__new__(cls, name, bases, attrs)


def make_conversion_dict(structure):
    """Build a default conversion dictionary for `structure`

    The default conversion mapping includes translations for any elements which
    are subclasses of :class:`SchemaDocument`.

    """
    conversions = {}
    for key, value in structure.iteritems():
        conversion = _make_single_conversion(value)
        if conversion:
            conversions[key] = conversion

    if conversions:
        return conversions


def _make_single_conversion(value):
    if isinstance(value, list):
        return _make_single_conversion(value[0])

    if isinstance(value, dict):
        return make_conversion_dict(value)

    if isinstance(value, type) and issubclass(value, SchemaDocument):
        return value

def find_shard_index(indexes):
    """Find the shard key and validate index properties.
    """
    shard_key_indexes = [index for index in indexes
                         if index.get('shard_key')]
    if not shard_key_indexes:
        return None

    if len(shard_key_indexes) > 1:
        raise SchemaError('There can only be one shard key per collection.')

    shard_index = shard_key_indexes[0]

    unique_indexes = [index for index in indexes if index.get('unique')]

    if len(unique_indexes) > 1:
        raise SchemaError(
            'A sharded collection may only have one unique index.')

    if unique_indexes and unique_indexes != shard_key_indexes:
        raise SchemaError(
            'Only the shard key may be used as a unique index on'
            ' a sharded collection.')

    return shard_index


class SchemaDocument(ConversionDict):
    """Base class for all documents with an enforced schema.

    This class also uses the structure information to wrap the values of
    embedded non-primatives.
    """

    __metaclass__ = SchemaMetaclass

    NONKEY_ATTRS = set(['collection', 'database', 'connection']) & \
        ConversionDict.NONKEY_ATTRS

    def __init__(self, *args, **kwargs):
        content = dict(*args, **kwargs)
        ConversionDict.__init__(self, content, self._conversions)

    def validate(self):
        validate_structure(self, self.structure)
        validate_required_fields(self, self.required_fields)


def validate_structure(body, structure):
    StructureWalker(validate_single_field).walk_dict(body, structure)


def validate_single_field(path, value, expected_type):
    """Validate a single field's value.

    This is the callback validator used by the :meth validate_structure: to
    check individual field values.
    """
    if not is_field_of_expected_type(value, expected_type):
        raise ValidationError(
            "Position {0!r} was declared to be {1}, but encountered value {2}"
            .format(path, expected_type, value))


def is_field_of_expected_type(value, expected_type):
    """Return ``True`` iff `value` meets the type description of `expected_type`.
    """
    if hasattr(expected_type, 'evaluate'):
        return expected_type.evaluate(value)

    return isinstance(value, expected_type)


def validate_required_fields(fields, required):
    """Ensure that all `required` fields are present in `fields`.
    """
    missing = required.difference(fields.keys())
    if missing:
        raise ValidationError(
            'Missing required field(s) {0}'.format(
                ','.join([repr(name) for name in missing])))


def validate_update_modifier(spec, structure):
    """Ensure that all operations are valid for the specified structure.
    """
    for modifier, args in spec.iteritems():
        validate_single_modifier(modifier, args, structure)


def validate_single_modifier(modifier, args, structure):
    """Validate a single update modifier.

    :param modifier: is a single update modifier (e.g. ``$set``).
    :param args: is the dictionary of arguments provided to the modifier.
    :param structure: Is the structure of the document that the
        modification should be applied to.  This is used to determine whether
        or not the modification is sane for this document.
    """
    if modifier == '$set':
        return validate_structure(args, structure)
    if modifier == '$unset':
        # TODO: Do not allow unsetting required fields or shard_key fields.
        return
    if modifier == '$rename':
        for old_name, new_name in args.iteritems():
            _validate_field_rename(old_name, new_name, structure)
        return

    for field, value in args.iteritems():
        _validate_field_modifier(modifier, structure[field], value)


def _validate_field_rename(old_name, new_name, structure):
    if structure[old_name] != structure[new_name]:
        raise ValidationError(
            'Cannot rename field of type {0} to field of type {1}'.format(
                repr(structure[old_name]), repr(structure[new_name])))


def _validate_field_modifier(modifier, field_type, value):
    # Since $pop, $pull, $pullAll and $bit cannot be used to break a schema
    # (i.e. they can't change the type or add data) adding validation for their
    # content is not a priority.  Perhaps at some point they'll be validated,
    # but for now just let them pass.
    null_validator = lambda field, value: None
    validator = {
        '$inc': _validate_inc_modifier,
        '$push': _validate_push_modifier,
        '$pushAll': _validate_push_all_modifier,
        '$addToSet': _validate_add_to_set_modifier,
        '$pop': null_validator,
        '$pull': null_validator,
        '$pullAll': null_validator,
        '$bit': null_validator,
    }.get(modifier)

    if not validator:
        raise ValidationError(
            'Encountered unknown update modifier {0}'.format(repr(modifier)))

    validator(field_type, value)


def _validate_inc_modifier(field_type, value):
    if field_type not in [int, long, float]:
        raise ValidationError(
            'Cannot increment non-numeric field of declared as {0}'.format(
                repr(field_type)))


def _make_single_value_array_modifier_validator(action_name):
    """Make and return a closure that is a validator for modifiers that add a
    single element to arrays.  I.e $push and $addToSet.
    """
    def _validator(field_type, value):
        if not isinstance(field_type, list):
            raise ValidationError(
                'Cannot {0} values onto non-array field of {1}'.format(
                    action_name, repr(field_type)))

        array_type = field_type[0]
        if not is_field_of_expected_type(value, array_type):
            raise ValidationError(
                'Cannot {0} value {1} onto array of {2}'.format(
                    action_name, repr(value), repr(array_type)))
    return _validator


_validate_push_modifier = _make_single_value_array_modifier_validator('push')
_validate_add_to_set_modifier = _make_single_value_array_modifier_validator('$addToSet')


def _validate_push_all_modifier(field_type, value):
    if not isinstance(value, list):
        raise ValidationError(
            'Cannot use modifier $pushAll with non-array argument {0}'.format(
                value))
    for subvalue in value:
        _validate_push_modifier(field_type, subvalue)
